package io.tswf.groovy.bettergroovy.asm.transmutation

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.At
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtExit
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtFieldCall
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtHead
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtMethodCall
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtReturn
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ClassReference
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.Constant
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.CreateLabel
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.CreateLabels
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.DependedAt
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.FieldValue
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.IsReturned
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.IsThrown
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.JumpingToCreatedLabelIfNotNull
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.LocalVariableValue
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.MethodDescription
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.Offset
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ReturnedValue
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ReturningConstantIfTrue
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ReturningIntoLocalVariable
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ReturningValueIfNotNull
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ReturningVoidIfTrue
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.Self
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ThrownException
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.Transmute
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.TransmutationInjectionContextReader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.TransmutationInjector
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.TransmutationMethodAtInsnInjector
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.InjectionOffset
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.InjectionPointSearchContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.MethodCallInsnType
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.TransmutationInjectionPointInsnFinder
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.TransmutationClassReader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.TransmutationMetadataReader
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.MissingTargetStrategy
import org.objectweb.asm.ClassReader
import org.objectweb.asm.ClassWriter
import org.objectweb.asm.commons.ClassRemapper
import org.objectweb.asm.commons.Remapper
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.InsnList
import org.objectweb.asm.tree.JumpInsnNode
import org.objectweb.asm.tree.LabelNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.VarInsnNode

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class App {
    static void blablabla() {
        try {
            println "q123"
        } catch (Throwable e) {
            throw new RuntimeException("Hello!", e)
        }
    }

    static ClassNode getTestClassNode() {
        def classReader = new ClassReader("io.tswf.groovy.bettergroovy.asm.transmutation.TestRemappingTargetClass")
        def classNode = new ClassNode()
        classReader.accept(classNode, ClassReader.SKIP_FRAMES)

        return classNode
    }

    static void saveTestClassNode(ClassNode classNode) {
        new File("output.class").setContent(classNode.toByteArray())
    }

    static void main(String[] args) {
//        testInjectionPoints()
        transmuteTest()
    }

    static void transmuteTest() {
        def transmutationsClass = TestTransmutationsContainer.toClassNode()
        def transmutations = TransmutationClassReader.readAllTransmutationsInClass(transmutationsClass)
        def targetClass = ClassLoader.contextClassloader().loadClass(transmutations.first().targetClass.className).toClassNode()
        TransmutationInjector.injectTransmutations(transmutations, targetClass)

        def targetClassBytes = targetClass.toByteArray(
//                ClassWriter.COMPUTE_MAXS
        )
        def targetClassOutputFile = new File("./target-class-output.class")
        targetClassOutputFile.bytes = targetClassBytes
    }

//    static void testInjectionPoints() {
//        def methodCallInjectionPoint = new MethodCallTransmutationInjectionPoint()
//        def testClass = TestRemappingTargetClass.toClassNode()
//
//        def pointAnnotation = testClass.methods
//                .find { it.name == "transmutatePoint" }
//                .visibleAnnotations
//                .find { it.desc == AtMethodCall.asmDescriptor }
//
//        def transmutateTargetMethod = testClass.methods.find { it.name == "remapMeForANullCheck" }
//
//        def context = new InjectionPointSearchContext(
//                transmutationTargetMethodNode: transmutateTargetMethod,
//                insnFromSearch: transmutateTargetMethod.instructions.first,
//                transmutationMethodMetadata: null,
//                injectionPointMarkerAnnotation: pointAnnotation
//        )
//
//        def result = methodCallInjectionPoint.findNextNodeNearInject(context)
//        println result
//
//        if (result instanceof MethodInsnNode) {
//            println result.name
//        }
//    }

    static void testClone() {
        def classNode = getTestClassNode().cloneClassNode()
        saveTestClassNode(classNode)
    }

    void virtualVoid1() {  }
    void virtualVoid2() {
        this.also {
            virtualVoid1()
            println 123
            privateVirtualVoid()
        }
    }

    private static void privateStaticVoid() {
        println "Hello World!"
    }

    private void privateVirtualVoid() {
        println "Hello World 2!"
    }

    static void testJump() {
        def classNode = getTestClassNode()

        def methodNode = classNode.methods.find {it.name == "remapMeForANullCheck" }
        def lastToStringInvocationInsn = methodNode.instructions.stream()
            .filter { it instanceof MethodInsnNode && it.name == "toString" }
            .toList().last()

        classNode.with {
            privateStaticVoid()
        }

        def jumpTargetLabel = new LabelNode()

        methodNode.instructions.insertAfter(lastToStringInvocationInsn, jumpTargetLabel)

        def newNodes = new InsnList()
        newNodes.add(new VarInsnNode(ALOAD, 0))
        newNodes.add(new JumpInsnNode(IFNULL, jumpTargetLabel))

        methodNode.instructions.insertBefore(lastToStringInvocationInsn, newNodes)

        saveTestClassNode(classNode)
    }

    static void testRemapper() {
        def classNode = getTestClassNode()

        def remappedClassNode = new ClassNode()
        def remapperClassVisitor = new ClassRemapper(remappedClassNode, new TestRemapper())
        classNode.accept(remapperClassVisitor)

        saveTestClassNode(remappedClassNode)
    }
}

@CompileStatic
class TestTransmutationsContainer {

    @Transmute(
            targetClass = @ClassReference(TestTransmutationTarget),
            targetMethod = @MethodDescription(name = "getSomeStringNotExisting", parameterTypeClasses = [ String ], returnTypeClass = String),
            onMissing = MissingTargetStrategy.CREATE_NEW
    )
    @AtExit
    @ReturningValueIfNotNull
    static String injectHelloWorld(String str) {
        return "123"
    }

    //TODO: Написать тесты, которые будут дампить байты классов и пытаться загружать их для самых основных случаев.
    @Transmute(
            targetClass = @ClassReference(TestTransmutationTarget),
            targetMethod = @MethodDescription(name = "getSomeString", descriptor = "(Ljava/lang/String;)V")
    )
//    @Offset(InjectionOffset.NO)
    @AtExit
//    @ReturningValueIfNotNull
//    @ReturningConstantIfTrue(value = @Constant(stringValue = "New Hello World!"))
//    @ReturningVoidIfTrue
//    @JumpingToCreatedLabelIfNotNull
//    @CreateLabels([
//            @CreateLabel(name = "beforeMarker", at = [ @At(atMethod = @AtMethodCall(owner = @ClassReference(MarkersClass), method = @MethodDescription(name = "someMarkerMethod", descriptor = "()V"), callType = MethodCallInsnType.STATIC)) ]),
//            @CreateLabel(name = "return", at = [ @At(atReturn = @AtReturn, offset = @Offset(InjectionOffset.BEFORE)) ])
//    ])
//    @ReturningIntoLocalVariable(localVariableIndex = 1, localVariableType = @ClassReference(String))
    @ReturningVoidIfTrue
    static boolean injectHelloWorld(
            String str,

            @IsReturned
            boolean isReturned,

//            @ReturnedValue
//            String oldReturnValue,

            @IsThrown
            boolean thrown,

            @Self
            Object self,

            @ThrownException
            Throwable exception,

            @LocalVariableValue(index = 1, type = @ClassReference(String))
            Object firstLocal,

            @FieldValue(name = "superSecretStaticStringField", type = @ClassReference(String))
            String staticStringFieldValue
    )
    {
        println "Hello, World! Arg: '$str'!"
//        return "beforeMarker"
//        return -1
        return true
    }
}

@CompileStatic
class TestTransmutationTarget {

    private static String superSecretStaticStringField

    void getSomeString(final String anotherString) {
//        println "Hello!"
        MarkersClass.someMarkerMethod()
//        println "Hello 2!"

        String helloWorldString = "Hello World String"

        if (Random.randomBoolean()) {
            throw new NullPointerException()
        }

        println "Hello World!"
    }
}

@CompileStatic
class MarkersClass {

    static void someMarkerMethod() { println "123" }

}

@CompileStatic
class TestRemappingTargetClass {
    public static Runnable blablabla

    @AtMethodCall(
            method = @MethodDescription(
                    name = "someStaticMethodToInvoke",
                    descriptor = "()V"
            ),
            owner = @ClassReference(TestRemappingTargetClass),
            callType = MethodCallInsnType.STATIC
    )
    static void transmutatePoint() {

    }

    static void remapMeForANullCheck(Object arg) {
        println "Hello null safe world!"

        arg.toString()

        someStaticMethodToInvoke()

        println "I am here!"
    }

    static void someStaticMethodToInvoke() {

    }
}

@CompileStatic
class TestRemapper extends Remapper {
    Map<String, String> mappings = new HashMap<String, String>().tap {
        put(Runnable.asmInternalName, "java/lang/Double")
    }

    @Override
    String map(String internalName) {
        return mappings[internalName] ?: super.map(internalName)
    }
}


@CompileStatic
class TestContainer {

    @Transmute (
            targetClass = @ClassReference(Object),
            targetMethod = @MethodDescription (
                    name = "getPlayerInventory",
                    parameterTypeClasses = [ String, String ],
                    returnTypeClass = String
            ),
            dependedAt = [
                @DependedAt([
                        @At(
                            atField = @AtFieldCall(name = "field", descriptor = "Ljava/lang/String;", owner = @ClassReference(Object)),
                            offset = @Offset(InjectionOffset.AFTER),
                            skip = 1
                        ),
                        @At(
                            atField = @AtFieldCall(name = "field", descriptor = "Ljava/lang/String;", owner = @ClassReference(Object)),
                            offset = @Offset(InjectionOffset.AFTER)
                        )
                ])
            ]
    )
    @ReturningValueIfNotNull
    @AtReturn
    @Offset(InjectionOffset.AFTER)
    static Object replaceNullReturnValue(String param1, String param2, @Self Object self, @ReturnedValue Object oldReturnValue) {
        if (oldReturnValue == null) {
            return UUID.randomUUIDString()
        }

        return null
    }

    @Transmute (
            targetMethod = @MethodDescription (
                    name = "setPlayerInventory",
                    parameterTypeClasses = [ String, String ],
                    returnType = @ClassReference(descriptor = "I")
            )
    )
    @ReturningConstantIfTrue(@Constant(intValue = 11))
    @AtReturn
    static boolean returnConstantIfTrue(String param1, String param2, @Self Object self, @ReturnedValue int oldReturnValue) {
        if (oldReturnValue < 10) {
            return true
        }

        return false
    }

}