package io.tswf.groovy.bettergroovy.asm.transmutation.exception

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodMetadata
import org.objectweb.asm.tree.ClassNode

@CompileStatic
class InvalidTransmutationTargetClassException extends RuntimeException {
    Collection<TransmutationMethodMetadata> transmutationMethodsMeta
    ClassNode targetClass

    InvalidTransmutationTargetClassException(Collection<TransmutationMethodMetadata> transmutationMethodsMeta, ClassNode targetClass) {
        super("There is no ${targetClass.javaName} in transmutations: ${System.lineSeparator()}${getFancyTransmutationsString(transmutationMethodsMeta)}")
        this.transmutationMethodsMeta = transmutationMethodsMeta
        this.targetClass = targetClass
    }

    static String getFancyTransmutationsString(Collection<TransmutationMethodMetadata> transmutationMethodsMeta) {
        def transmutationsListAsFancyString = transmutationMethodsMeta
            .groupBy { it.transmutationOwner.javaName }
            .entrySet()
                .stream()
                    .map {
                        def transmutationsInCurrentClassFancyString = it.value.stream()
                            .map { tr -> tr.transmutationMethod.javaFullName }
                            .joinToString(",${System.lineSeparator()}")

                        "${it.key} methods: ${transmutationsInCurrentClassFancyString}"
                    }
                    .joinToString(System.lineSeparator())


        return transmutationsListAsFancyString
    }
}

