package io.tswf.groovy.bettergroovy.asm.transmutation.injection

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.TransmutationInjectionPointInsnFinder
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodMetadata
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class TransmutationInjectionContextReader {

    static TransmutationInjectionToMethodContext readInjectionContext(
            TransmutationMethodMetadata transmutationMetadata,
            ClassNode targetClass,
            MethodNode targetMethod
    ) {
        def insns = transmutationMetadata.injectionPoints.stream()
            .flatMap {
                TransmutationInjectionPointInsnFinder.findAllInsnsAfterWhichInject(it, targetMethod)
                    .stream()
            }
            .toList()

        def result = new TransmutationInjectionToMethodContext(
                transmutationMetadata: transmutationMetadata,
                targetMethodOwner: targetClass,
                targetMethod: targetMethod
        )

        for (def insn : insns) {
            def insnInjectionContext = new TransmutationInjectionToMethodAtInsnContext(
                    startInsn: insn,
                    allPointsContext: result

            )

            result.perInsnContexts << insnInjectionContext
        }

        return result
    }
}
