package io.tswf.groovy.bettergroovy.asm.transmutation.injection

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationTargetClassException
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodMetadata
import org.objectweb.asm.tree.ClassNode

@CompileStatic
class TransmutationInjector {

    static void injectTransmutations(
            Collection<TransmutationMethodMetadata> transmutationMethodsMeta,
            ClassNode targetClass
    ) {
        def targetClassType = targetClass.asAsmType()
        def matchingTransmutationsMeta = transmutationMethodsMeta.filter { targetClassType in it.targetClass }

        if (matchingTransmutationsMeta.isEmpty()) {
            throw new InvalidTransmutationTargetClassException(transmutationMethodsMeta, targetClass)
        }

        for (def transmutationMeta : matchingTransmutationsMeta) {
            def matchingMethod = TransmutationTargetMethodsFinder.findMatchingTargetMethodForTransmutation(transmutationMeta, targetClass)

            // Если тут еще не вылетело исключения, но целевой метод все равно не найден, то значит по логике надо его пропустить.
            if (matchingMethod == null) {
                continue
            }

            def injectionContext = TransmutationInjectionContextReader.readInjectionContext(transmutationMeta, targetClass, matchingMethod)
            for (def insnInjectionContext : injectionContext.perInsnContexts) {
                TransmutationMethodAtInsnInjector.injectTransmutation(insnInjectionContext)
            }
        }
    }
}
