package io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InjectionPointCantBeReachedException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationPreInjectionHook
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.TransmutationInjectionPointInsnFinder
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodInjectionPointMetadata
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor
import org.objectweb.asm.tree.LabelNode

import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.NEW_CREATED_LABELS

@CompileStatic
class CreateLabelsPreInjectionHook implements ITransmutationPreInjectionHook {

    List<CreateLabelMetadata> newLabelsMetadata

    CreateLabelsPreInjectionHook(List<CreateLabelMetadata> newLabelsMetadata) {
        this.newLabelsMetadata = newLabelsMetadata
    }

    @Override
    void handleTransmutationPreInjection(TransmutationInjectionToMethodAtInsnContext context) {
        def targetMethod = context.allPointsContext.targetMethod

        // Все лейблы должны создаваться только один раз на трансмутацию, чтобы не загаживать байткод.
        def newCreatedLabelsMap = context.allPointsContext.customPayload.computeIfAbsent(NEW_CREATED_LABELS) {
            new HashMap<String, LabelNode>()
        }

        // Хранятся они в кастомных данных контекста трансмутации как карта "Имя -> LabelNode"
        // Далее любой обработчик сможет достать эту карту и использовать ее.
        for (def newLabelMeta : newLabelsMetadata) {
            newCreatedLabelsMap.computeIfAbsent(newLabelMeta.name) {
                def insnBeforeCreate = TransmutationInjectionPointInsnFinder.findInsnAfterWhichInject(newLabelMeta.injectionPointMetadata, context.allPointsContext.targetMethod)
                def checkTargetInsnExists = targetMethod.instructions.contains(insnBeforeCreate)
                if (checkTargetInsnExists) {
                    def newCreatedLabel = new LabelNode()
                    targetMethod.instructions.insertBefore(insnBeforeCreate, newCreatedLabel)
                    return newCreatedLabel
                } else if (newLabelMeta.required) {
                    throw new InjectionPointCantBeReachedException("Instruction needed for label '${newLabelMeta.name}' creation was not found in ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName}")
                }

                return null
            }
        }
    }

    @CheckedMapConstructor
    @CompileStatic
    static class CreateLabelMetadata {
        TransmutationMethodInjectionPointMetadata injectionPointMetadata
        String name
        boolean required
    }
}
