package io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationPostInjectionHook
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import org.objectweb.asm.Opcodes
import org.objectweb.asm.Type
import org.objectweb.asm.tree.VarInsnNode

import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.NEEDS_TO_LOAD_RETURN_LOCAL_VAR_KEY
import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.NEEDS_TO_LOAD_THROWN_EXCEPTION_LOCAL_VAR_KEY
import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.RETURN_LOCAL_VAR_INDEX_KEY
import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.THROWN_EXCEPTION_LOCAL_VAR_KEY
import static org.objectweb.asm.Opcodes.ALOAD
import static org.objectweb.asm.Opcodes.ATHROW
import static org.objectweb.asm.Opcodes.ILOAD

@CompileStatic
class LoadThrownExceptionPostInjectionHook implements ITransmutationPostInjectionHook {

    @Override
    void handleTransmutationPostInjection(TransmutationInjectionToMethodAtInsnContext context) {
        if (context.startInsn.opcode == ATHROW && context.customPayload.getByKey(NEEDS_TO_LOAD_THROWN_EXCEPTION_LOCAL_VAR_KEY)) {
            def targetMethod = context.allPointsContext.targetMethod

            def thrownExceptionLocalVarIndex = (int) context.allPointsContext.customPayload.getByKey(THROWN_EXCEPTION_LOCAL_VAR_KEY)
                .ensureNotNull { new IllegalStateException("There is no thrown exception local var storage index for method ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName}") }

            context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, thrownExceptionLocalVarIndex)
        }
    }
}
