package io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationPreInjectionHook
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import org.objectweb.asm.Opcodes
import org.objectweb.asm.Type
import org.objectweb.asm.tree.VarInsnNode

import static org.objectweb.asm.Opcodes.*
import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.*

@CompileStatic
class StoreReturnValuePreInjectionHook implements ITransmutationPreInjectionHook {

    @Override
    void handleTransmutationPreInjection(TransmutationInjectionToMethodAtInsnContext context) {
        if (Opcodes.isOpcodeReturn(context.startInsn.opcode)) {
            def targetMethod = context.allPointsContext.targetMethod
            def targetMethodReturnType = targetMethod.returnType
            if (targetMethodReturnType == Type.VOID_TYPE) {
                throw new InvalidTransmutationException("Can't store void return type in local variable for method ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName}")
            }

            // Переменная для хранения результата может быть одна на всех для удобства
            def returnLocalVarIndex = context.allPointsContext.customPayload.computeIfAbsent(RETURN_LOCAL_VAR_INDEX_KEY) {
                targetMethod.createNewLocal(targetMethodReturnType, "return", true)
            }

            // По-умолчанию считаем, что раз положили значение со стека в переменную, то туда его в конце и стоит вернуть.
            context.customPayload.putByKey(NEEDS_TO_LOAD_RETURN_LOCAL_VAR_KEY, true)

            context.insnsToInsertToCurrent << new VarInsnNode(targetMethodReturnType.getOpcode(ISTORE), returnLocalVarIndex)
        }
    }
}
