package io.tswf.groovy.bettergroovy.asm.transmutation.injection.missing

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.IMissingTransmutationTargetMethodHandleExceptionBypassMarker
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.MissingTransmutationTargetMethodHandleException
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.MissingTargetStrategy
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodMetadata
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class TransmutationMissingTargetHandler {

    @Lazy
    static Map<MissingTargetStrategy, IMissingTargetStrategyLogic> missingTargetStrategies = ServiceLoader.load(IMissingTargetStrategyLogic)
        .toMapBy { it.enumValue }
        .tap { strategies ->
            def missingKeys = MissingTargetStrategy.values()
                .stream()
                    .filter { !strategies.containsKey(it) }
                    .toSet()

            if (missingKeys.isNotNullOrEmpty()) {
                throw new RuntimeException("There is no logic found for transmutation strategies: ${missingKeys}!")
            }
        }

    static MethodNode handleMissingTargetAndTryToPass(TransmutationMethodMetadata transmutationMethodMeta, ClassNode targetClass) {
        try {
            def strategy = missingTargetStrategies[transmutationMethodMeta.targetMissingStrategy]
                    .ensureNotNull { new RuntimeException("There is no logic found for transmutation missing target strategy ${transmutationMethodMeta.targetMissingStrategy}") }

            def missingTargetHandleContext = new MissingTargetHandleContext(
                    targetClass: targetClass,
                    transmutationMethodMeta: transmutationMethodMeta
            )

            strategy.handleMissingTarget(missingTargetHandleContext)

            return missingTargetHandleContext.newTargetMethod
        } catch (Throwable e) {
            if (e instanceof IMissingTransmutationTargetMethodHandleExceptionBypassMarker) {
                throw e
            }

            throw new MissingTransmutationTargetMethodHandleException(e, transmutationMethodMeta, targetClass)
        }
    }
}
