package io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.IsThrown
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationMethodCustomParameterLoader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnNode

import static org.objectweb.asm.Opcodes.*

@AutoService(ITransmutationMethodCustomParameterLoader)
@CompileStatic
class IsThrownCustomParameterLoader implements ITransmutationMethodCustomParameterLoader {

    @Override
    void loadCustomParameter(TransmutationInjectionToMethodAtInsnContext context, int parameterIndex, Type parameterType) {
        if (parameterType != Type.BOOLEAN_TYPE) {
            throw new InvalidTransmutationException("The annotated with @${IsThrown.simpleName} ${parameterIndex}'st parameter of ${context.allPointsContext.transmutationMetadata.transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName} type must be a primitive boolean, but was ${parameterType.className}")
        }

        context.insnsToInsertToCurrent << new InsnNode(context.startInsn.getOpcode() == ATHROW ? ICONST_1 : ICONST_0)
    }
}
