package io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.IsReturned
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.LocalVariableValue
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.NoLocalVariableFoundException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationMethodCustomParameterLoader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import org.objectweb.asm.Type
import org.objectweb.asm.tree.VarInsnNode

import static org.objectweb.asm.Opcodes.ILOAD

@CompileStatic
class LocalVariableCustomParameterLoader implements ITransmutationMethodCustomParameterLoader {

    int localVariableIndex
    Type localVariableType

    LocalVariableCustomParameterLoader(int localVariableIndex, Type localVariableType) {
        this.localVariableIndex = localVariableIndex
        this.localVariableType = localVariableType
    }

    @Override
    void loadCustomParameter(TransmutationInjectionToMethodAtInsnContext context, int parameterIndex, Type parameterType) {
        def targetMethod = context.allPointsContext.targetMethod
        def targetMethodLocals = targetMethod.localVariables

        def matchingLocalVar = targetMethodLocals
                .find { it.index == localVariableIndex && it.desc == localVariableType.descriptor }
                // TODO: В целом тут можно добавить загрузку какого-то фаллбэчного значения и варнинг в логи при определенной стратегии обработки ошибок
                .ensureNotNull { new NoLocalVariableFoundException("There is no local variable found with type '${localVariableType.descriptor}' and local index '${localVariableIndex}'. The list of available locals here: ${targetMethodLocals.map { "${it.index}: ${it.desc} (name: '${it.name}')" }}") }

        if (parameterType != localVariableType) {
            // Надо оставить возможность положить в параметр тип, который не доступен из класса трансмутации
            if (parameterType == Object.asAsmType() && !localVariableType.isPrimitive()) {
                // Nope!
            } else {
                throw new InvalidTransmutationException("The annotated with @${LocalVariableValue.simpleName} ${parameterIndex}'st parameter of ${context.allPointsContext.transmutationMetadata.transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName} type must be a ${localVariableType.className}, but was ${parameterType.className}")
            }
        }

        context.insnsToInsertToCurrent << new VarInsnNode(localVariableType.getOpcode(ILOAD), matchingLocalVar.index)
    }
}
