package io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ThrownException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationMethodCustomParameterLoader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.VarInsnNode

import static org.objectweb.asm.Opcodes.*
import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.*

@CompileStatic
class ThrownExceptionCustomParameterLoader implements ITransmutationMethodCustomParameterLoader {

    @Override
    void loadCustomParameter(TransmutationInjectionToMethodAtInsnContext context, int parameterIndex, Type parameterType) {
        if (parameterType != Throwable.asAsmType()) {
            throw new InvalidTransmutationException("The annotated with @${ThrownException} ${parameterIndex}'st parameter of ${context.allPointsContext.transmutationMetadata.transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName} type must be a Throwable, but was ${parameterType.className}")
        }

        if (context.startInsn.opcode == ATHROW) {
            def thrownExceptionStorageLocalVarIndex = (Integer) context.allPointsContext.customPayload.getByKey(THROWN_EXCEPTION_LOCAL_VAR_KEY)
                .ensureNotNull { new InvalidTransmutationException("There is no thrown exception local var storage found!") }

            context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, thrownExceptionStorageLocalVarIndex)
        } else {
            context.insnsToInsertToCurrent << new InsnNode(ACONST_NULL)
        }
    }
}
