package io.tswf.groovy.bettergroovy.asm.transmutation.injection.points

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtConstant
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationDescriptionException
import io.tswf.groovy.bettergroovy.asm.transmutation.util.ConstantUtil
import org.objectweb.asm.tree.AbstractInsnNode
import org.objectweb.asm.tree.AnnotationNode
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.IntInsnNode
import org.objectweb.asm.tree.LdcInsnNode

import static org.objectweb.asm.Opcodes.*

@AutoService(ITransmutationInjectionPoint)
@CompileStatic
class ConstantTransmutationInjectionPoint implements ITransmutationInjectionPoint{

    Object requestedConstantValue

    Map<String, Map<Integer, Object>> descriptorToConstInsnsMap = new HashMap<String, Map<Integer, Object>>()
        .also {
            def iconstMap = new HashMap<Integer, Object>().also {
                it.put(ICONST_M1, -1)
                it.put(ICONST_0, 0)
                it.put(ICONST_1, 1)
                it.put(ICONST_2, 2)
                it.put(ICONST_3, 3)
                it.put(ICONST_4, 4)
                it.put(ICONST_5, 5)
            }

            def lconstMap = new HashMap<Integer, Object>().also {
                it.put(LCONST_0, 0l)
                it.put(LCONST_1, 1l)
            }

            def fconstMap= new HashMap<Integer, Object>().also {
                it.put(FCONST_0, 0f)
                it.put(FCONST_1, 1f)
            }
            def dconstMap= new HashMap<Integer, Object>().also {
                it.put(DCONST_0, 0d)
                it.put(DCONST_1, 1d)
            }

            it.put(char.asmDescriptor, iconstMap)
            it.put(Character.asmDescriptor, iconstMap)

            it.put(boolean.asmDescriptor, iconstMap)
            it.put(Boolean.asmDescriptor, iconstMap)

            it.put(byte.asmDescriptor, iconstMap)
            it.put(Byte.asmDescriptor, iconstMap)

            it.put(short.asmDescriptor, iconstMap)
            it.put(Short.asmDescriptor, iconstMap)

            it.put(int.asmDescriptor, iconstMap)
            it.put(Integer.asmDescriptor, iconstMap)

            it.put(long.asmDescriptor, lconstMap)
            it.put(Long.asmDescriptor, lconstMap)

            it.put(float.asmDescriptor, fconstMap)
            it.put(Float.asmDescriptor, fconstMap)

            it.put(double.asmDescriptor, dconstMap)
            it.put(Double.asmDescriptor, dconstMap)
        }

    ConstantTransmutationInjectionPoint(Object requestedConstantValue) {
        this.requestedConstantValue = requestedConstantValue
    }

    @Override
    AbstractInsnNode findNextNodeNearInject(InjectionPointSearchContext context) {
        def insnCandidate = context.insnFromSearch
        def constInsnsOpcodes = descriptorToConstInsnsMap[requestedConstantValue.class.asmDescriptor]

        while (insnCandidate != null) {
            if (insnCandidate instanceof LdcInsnNode) {
                def constantValue = insnCandidate.cst

                if (constantValue == requestedConstantValue && constantValue.class == requestedConstantValue.class) {
                    return insnCandidate
                }
            } else if (insnCandidate instanceof InsnNode && constInsnsOpcodes != null) {
                def constantValue = constInsnsOpcodes[insnCandidate.opcode]

                if (constantValue == requestedConstantValue) {
                    return insnCandidate
                }
            } else if (insnCandidate instanceof IntInsnNode) {
                if (insnCandidate.opcode == BIPUSH || insnCandidate.opcode == SIPUSH) {
                    def constantValue = insnCandidate.operand
                    if (Objects.equals(requestedConstantValue, constantValue)) {
                        return insnCandidate
                    }
                }
            }

            insnCandidate = insnCandidate.next
        }

        return null
    }
}
