package io.tswf.groovy.bettergroovy.asm.transmutation.injection.points

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import org.objectweb.asm.Type
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.AbstractInsnNode
import org.objectweb.asm.tree.MethodInsnNode

import static org.objectweb.asm.Opcodes.*

@CompileStatic
@AutoService(ITransmutationInjectionPoint)
class MethodCallTransmutationInjectionPoint implements ITransmutationInjectionPoint {

    Type callMethodOwner
    Method callMethod
    MethodCallInsnType callType

    MethodCallTransmutationInjectionPoint(Type callMethodOwner, Method callMethod, MethodCallInsnType callType) {
        this.callMethodOwner = callMethodOwner
        this.callMethod = callMethod
        this.callType = callType
    }

    @Override
    AbstractInsnNode findNextNodeNearInject(
            InjectionPointSearchContext context
    ) {
        def insnOpcode = callType.opcode

        return context.insnFromSearch.asIterable()
            .stream()
                .peek { println it }
                .filter {insnCandidate ->
                    if (insnCandidate instanceof MethodInsnNode) {
                        if (insnCandidate.opcode != insnOpcode)
                            return false

                        if (insnCandidate.name != callMethod.name)
                            return false

                        if (insnCandidate.desc != callMethod.descriptor)
                            return false

                        if (insnCandidate.owner != callMethodOwner.internalName)
                            return false

                        return true
                    }

                    return false
                }
                .firstOrNull()
    }
}

@CompileStatic
enum MethodCallInsnType {
    VIRTUAL(INVOKEVIRTUAL),
    STATIC(INVOKESTATIC),
    SPECIAL(INVOKESPECIAL),
    INTERFACE(INVOKEINTERFACE)
    ;

    int opcode

    MethodCallInsnType(int opcode) {
        this.opcode = opcode
    }
}