package io.tswf.groovy.bettergroovy.asm.transmutation.injection.points

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodInjectionPointMetadata
import io.tswf.groovy.bettergroovy.asm.transmutation.util.InjectionInsnOffsetUtil
import org.objectweb.asm.tree.AbstractInsnNode
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class TransmutationInjectionPointInsnFinder {

    static List<AbstractInsnNode> findAllInsnsAfterWhichInject(
            TransmutationMethodInjectionPointMetadata injectionPointMetadata,
            MethodNode targetMethod,
            AbstractInsnNode startInsn = targetMethod.instructions.first
    ) {
        def currentInsn = startInsn
        def result = new ArrayList<AbstractInsnNode>()

        while (currentInsn != null) {
            def foundInsn = findInsnAfterWhichInject(injectionPointMetadata, targetMethod, currentInsn)
            if (foundInsn != null) {
                result << foundInsn
                currentInsn = foundInsn.next
            } else {
                break
            }
        }

        return result
    }

    static AbstractInsnNode findInsnAfterWhichInject(
            TransmutationMethodInjectionPointMetadata injectionPointMetadata,
            MethodNode targetMethod,
            AbstractInsnNode startInsn = targetMethod.instructions.first
    ) {
        def pointsMetadataChain = new ArrayDeque<TransmutationMethodInjectionPointMetadata>()

        // Используем tap, чтобы убрать лишние переменные из поля зрения
        injectionPointMetadata.tap {
            def next = it
            while (next != null) {
                pointsMetadataChain.offerFirst(next)
                next = next.parent
            }
        }

        def currentInsn = startInsn
        def currentInjectionPointMeta = pointsMetadataChain.pollLast()

        // Начинаем искать пока не занулилась или инструкция (это значит, что где-то что-то не нашли или мета injection point'а, что означает, что мы прошли всю цепочку)
        while (currentInjectionPointMeta != null && currentInsn != null) {
            def skipCount = currentInjectionPointMeta.skip
            def foundInsn = null

            // Учитываем skip и пропускаем значения в том количестве, которое нужно
            do {
                def insnSearchContext = new InjectionPointSearchContext(
                        transmutationTargetMethodNode: targetMethod,
                        insnFromSearch: currentInsn

                )

                foundInsn = currentInjectionPointMeta.injectionPoint.findNextNodeNearInject(insnSearchContext)

                if (foundInsn == null) {
                    return null
                }

                skipCount --
            }
            while (skipCount >= 0)

            // Применяем смещение для найденной инструкции
            currentInsn = foundInsn?.with { InjectionInsnOffsetUtil.applyInsnOffset(targetMethod, foundInsn, currentInjectionPointMeta.injectionOffset) }

            // Переходим к следующей по очереди мете injection point'а
            currentInjectionPointMeta = pointsMetadataChain.pollLast()
        }

        // В результате инструкция, которая была найдена последней - и есть наша.
        return currentInsn
    }

}
