package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.IsReturned
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationDescriptionException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter.IsReturnedCustomParameterLoader
import org.objectweb.asm.Type

@CompileStatic
@AutoService(ICustomParameterLoaderReader)
class IsReturnedCustomParameterLoaderReader implements ICustomParameterLoaderReader {

    @Override
    void handleLoaderSearch(CustomParameterLoaderReadingContext readingContext) {
        def transmutationMethod = readingContext.transmutationMetadata.transmutationMethod

        if (transmutationMethod.hasParameterAnnotation(readingContext.parameterIndex, IsReturned)) {
            def parameterType = transmutationMethod.parameterTypes[readingContext.parameterIndex]

            if (parameterType != Type.BOOLEAN_TYPE) {
                throw new InvalidTransmutationDescriptionException("The ${readingContext.parameterIndex + 1}'st parameter of method ${transmutationMethod.javaFullName} marked by @${IsReturned.simpleName} annotation, but it's type is not primitive boolean (was ${parameterType.descriptor}). It is not allowed!")
            }

            readingContext.transmutationMetadata.customParameterLoaders << new IsReturnedCustomParameterLoader()
            readingContext.completed = true
        }
    }
}
