package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.FieldValue
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter.LoadFieldCustomParameterLoader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.remap.TransmutationRemapper
import io.tswf.groovy.bettergroovy.asm.transmutation.util.ClassReferenceUtil
import org.objectweb.asm.tree.AnnotationNode

@AutoService(ICustomParameterLoaderReader)
@CompileStatic
class LoadFieldCustomParameterLoaderReader implements ICustomParameterLoaderReader {

    @Override
    void handleLoaderSearch(CustomParameterLoaderReadingContext readingContext) {
        def fieldAnnotation = readingContext.transmutationMetadata
                .transmutationMethod
                .findParameterAnnotationByType(readingContext.parameterIndex, FieldValue)

        if (fieldAnnotation != null) {
            def fieldName = fieldAnnotation.getValueByName("name", String)
            def fieldTypeRef = fieldAnnotation.getValueByName("type", AnnotationNode)
            def fieldType = ClassReferenceUtil.getTypeFromClassRefAnnotation(fieldTypeRef)

            def (mappedFieldName, mappedFieldType) = TransmutationRemapper.mapField(readingContext.transmutationMetadata.targetClass, fieldName, fieldType)

            readingContext.transmutationMetadata.customParameterLoaders << new LoadFieldCustomParameterLoader(mappedFieldName, mappedFieldType)
            readingContext.completed = true
        }
    }
}
