package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ReturnedValue
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.ThrownException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationDescriptionException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter.ReturnedValueCustomParameterLoader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.parameter.ThrownExceptionCustomParameterLoader

@AutoService(ICustomParameterLoaderReader)
@CompileStatic
class ThrownExceptionCustomParameterLoaderReader implements ICustomParameterLoaderReader {

    @Override
    void handleLoaderSearch(CustomParameterLoaderReadingContext readingContext) {
        if (readingContext.transmutationMetadata.transmutationMethod.hasParameterAnnotation(readingContext.parameterIndex, ThrownException)) {
            if (readingContext.parameterType.descriptor != Throwable.asmDescriptor) {
                throw new InvalidTransmutationDescriptionException("@${ThrownException.simpleName} annotated parameter's type in ${readingContext.transmutationMetadata.transmutationMethod.javaFullName} of ${readingContext.transmutationMetadata.transmutationOwner.javaName} must be a ${Throwable.name}, not a ${readingContext.parameterType.className}!")
            }

            readingContext.transmutationMetadata.customParameterLoaders << new ThrownExceptionCustomParameterLoader()
            readingContext.completed = true
        }
    }
}
