package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.hook

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.CreateLabel
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.CreateLabels
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationDescriptionException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.CreateLabelsPreInjectionHook
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.IPreInjectionHooksReader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.PreInjectionHooksReadingContext
import io.tswf.groovy.bettergroovy.asm.transmutation.util.CompositeAtUtil
import org.objectweb.asm.tree.AnnotationNode

@AutoService(IPreInjectionHooksReader)
@CompileStatic
class CreateLabelsHookReader implements IPreInjectionHooksReader {

    @Override
    void handlePreInjectionHooksReading(PreInjectionHooksReadingContext readingContext) {
        // Ищем множественную аннотацию и единственную прямо над методом
        def transmutationMethod = readingContext.transmutationMetadata.transmutationMethod
        def createLabelsAnnotation = transmutationMethod.findAnnotationByType(CreateLabels)

        def createLabelOnMethodAnnotation = transmutationMethod.findAnnotationByType(CreateLabel)

        // Если есть хотя бы одна, то пробуем распарсить
        if (createLabelsAnnotation != null || createLabelOnMethodAnnotation != null) {
            def listOfCreateLabelAnnotations = new ArrayList<AnnotationNode>()

            // Добавляем все аннотации из множественной версии, если она есть
            createLabelsAnnotation?.let {
                listOfCreateLabelAnnotations.addAll(
                        createLabelsAnnotation.getValueByName("value", AnnotationNode[])
                                .toList()
                )
            }

            // Добавляем аннотацию напрямую над методом, если она есть
            createLabelOnMethodAnnotation?.let {
                listOfCreateLabelAnnotations << it
            }

            // Читаем все аннотации и переводим их в формат метадаты для создания лейблов
            def newLabelMetadataList = listOfCreateLabelAnnotations.map {
                def required = it.getValueByName("required", Boolean) { false }
                def labelName = it.getValueByName("name", String)

                def injectionPointMetadataList = CompositeAtUtil.findInjectionPointsComposite(it)

                if (injectionPointMetadataList.size() > 1) {
                    throw new InvalidTransmutationDescriptionException("There is multiple injection points declared for label named '${labelName}' in @${CreateLabel.simpleName} annotation!")
                }

                if (injectionPointMetadataList.isEmpty()) {
                    throw new InvalidTransmutationDescriptionException("There is no injection points declared for label named '${labelName}' in @${CreateLabel.simpleName} annotation!")
                }

                def newLabelMetadata = new CreateLabelsPreInjectionHook.CreateLabelMetadata(
                        injectionPointMetadata: injectionPointMetadataList.first(),
                        name: labelName,
                        required: required
                )

                return newLabelMetadata
            }

            // Добавляем хук в трансмутацию
            readingContext.transmutationMetadata.preInjectionHooks << new CreateLabelsPreInjectionHook(newLabelMetadataList)
        }
    }
}
