package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.points

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtFieldCall
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationDescriptionException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.FieldCallTransmutationInjectionPoint
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.FieldCallType
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.IInjectionPointReader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.InjectionPointReadingContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.remap.TransmutationRemapper
import io.tswf.groovy.bettergroovy.asm.transmutation.util.ClassReferenceUtil
import org.objectweb.asm.Type
import org.objectweb.asm.tree.AnnotationNode

@AutoService(IInjectionPointReader)
@CompileStatic
class FieldCallTransmutationInjectionPointReader implements IInjectionPointReader {

    public static final String ATF_TYPE_PARAMETER_NAME = "callType"
    public static final String ATF_NAME_PARAMETER_NAME = "name"
    public static final String ATF_DESC_PARAMETER_NAME = "descriptor"
    public static final String ATF_OWNER_CLASS_PARAMETER_NAME = "owner"
    public static final String ATF_FIELD_TYPE_PARAMETER_NAME = "type"

    @Override
    void handleInjectionPointReading(InjectionPointReadingContext context) {
        def atFieldAnnotation = context.annotationNodeFromRead
        def callType = (atFieldAnnotation.findValueByName(ATF_TYPE_PARAMETER_NAME, FieldCallType)
                ?: FieldCallType.ANY)

        def rawFieldOwner = ClassReferenceUtil.getTypeFromClassRefAnnotation(
                atFieldAnnotation.getValueByName(ATF_OWNER_CLASS_PARAMETER_NAME, AnnotationNode)
        )

        def rawFieldName = atFieldAnnotation.getValueByName(ATF_NAME_PARAMETER_NAME, String)
        def rawFieldDescriptor = atFieldAnnotation.findValueByName(ATF_DESC_PARAMETER_NAME, String)
        def rawFieldType = atFieldAnnotation.findValueByName(ATF_FIELD_TYPE_PARAMETER_NAME, AnnotationNode)
            ?.with {
                ClassReferenceUtil.getTypeFromClassRefAnnotation(it)
            }

        if (rawFieldDescriptor == null && rawFieldType == null) {
            throw new InvalidTransmutationDescriptionException("There is no field type! Please, select one from [ ${ATF_DESC_PARAMETER_NAME}, ${ATF_FIELD_TYPE_PARAMETER_NAME} ] parameters!")
        }

        if (rawFieldDescriptor != null && rawFieldType != null) {
            throw new InvalidTransmutationDescriptionException("There is both field type parameters found! Please, choose only one from [ ${ATF_DESC_PARAMETER_NAME}, ${ATF_FIELD_TYPE_PARAMETER_NAME} ] parameters!")
        }

        def mappedFieldOwner = TransmutationRemapper.mapType(rawFieldOwner)

        def unmappedFieldType = Type.getType(rawFieldDescriptor ?: rawFieldType?.descriptor)
        def (mappedFieldName, mappedFieldType) = TransmutationRemapper.mapField(rawFieldOwner, rawFieldName, unmappedFieldType)

        context.result = new FieldCallTransmutationInjectionPoint(mappedFieldOwner, mappedFieldName, mappedFieldType.descriptor, callType)
    }

    @Override
    Class getMarkerAnnotationClass() {
        return AtFieldCall
    }
}
