package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.points

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.AtMethodCall
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.MethodCallInsnType
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.MethodCallTransmutationInjectionPoint
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.IInjectionPointReader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.InjectionPointReadingContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.remap.TransmutationRemapper
import io.tswf.groovy.bettergroovy.asm.transmutation.util.ClassReferenceUtil
import io.tswf.groovy.bettergroovy.asm.transmutation.util.MethodDescriptionUtil
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
@AutoService(IInjectionPointReader)
class MethodCallTransmutationInjectionPointReader implements IInjectionPointReader {

    public static final String ATM_OWNER_CLASS_PARAMETER_NAME = "owner"
    public static final String ATM_TYPE_PARAMETER_NAME = "callType"
    public static final String ATM_METHOD_DESCRIPTION_PARAMETER_NAME = "method"

    @Override
    void handleInjectionPointReading(InjectionPointReadingContext context) {
        def atMethodAnnotation = context.annotationNodeFromRead

        def rawCallMethod = MethodDescriptionUtil.findMethodFromAnnotation(
                context.annotationNodeFromRead.getValueByName(ATM_METHOD_DESCRIPTION_PARAMETER_NAME, AnnotationNode)
        )

        def rawCallMethodOwner = ClassReferenceUtil.getTypeFromClassRefAnnotation(
                atMethodAnnotation.getValueByName(ATM_OWNER_CLASS_PARAMETER_NAME, AnnotationNode)
        )

        def callType = atMethodAnnotation.findValueByName(ATM_TYPE_PARAMETER_NAME, MethodCallInsnType) ?: MethodCallInsnType.VIRTUAL

        def mappedCallMethodOwner = TransmutationRemapper.mapType(rawCallMethodOwner)
        def mappedCallMethod = TransmutationRemapper.mapMethod(rawCallMethodOwner, rawCallMethod)

        context.result = new MethodCallTransmutationInjectionPoint(mappedCallMethodOwner, mappedCallMethod, callType)
    }

    @Override
    Class getMarkerAnnotationClass() {
        return AtMethodCall
    }
}
