
package io.tswf.groovy.bettergroovy.asm.transmutation.injection.returnvalue

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.groovy.bettergroovy.asm.util.ConstantInsnUtil
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodNode

import static org.objectweb.asm.Opcodes.IRETURN

@CompileStatic
class ReturnConstantIfTransmutationTrueReturnValueProcessor extends AbstractIfTrueTransmutationTrueReturnValueProcessor {

    Object constantValue

    ReturnConstantIfTransmutationTrueReturnValueProcessor(Object constantValue) {
        this.constantValue = constantValue
    }

    @Override
    void validateTarget(MethodNode targetMethod) {
        def targetMethodReturnType = targetMethod.returnType
        if (targetMethodReturnType == Type.VOID_TYPE) {
            throw new InvalidTransmutationException("Can't return contant type from void method ${targetMethod.javaFullName}!")
        }
    }

    @Override
    void addReturnInsns(TransmutationInjectionToMethodAtInsnContext context) {
        if (constantValue == null) {
            throw new InvalidTransmutationException("The constant value can not be null!")
        }

        def constantValueType = constantValue.class.asAsmType().unboxed

        def targetMethod = context.allPointsContext.targetMethod
        def targetMethodReturnType = targetMethod.returnType

        if (targetMethodReturnType.descriptor != constantValueType.descriptor) {
            throw new InvalidTransmutationException("The constant value '${constantValue.toString()}' with type descriptor '${constantValueType.descriptor}' does not matches with target method's ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName} return type: ${targetMethodReturnType.descriptor}!")
        }

        context.insnsToInsertToCurrent << ConstantInsnUtil.instance.getConstantInsn(constantValue)
        context.insnsToInsertToCurrent << new InsnNode(context.allPointsContext.targetMethod.returnType.getOpcode(IRETURN))
    }
}
