package io.tswf.groovy.bettergroovy.asm.transmutation.util

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationDescriptionException
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
class ConstantUtil {

    public static final Collection<String> CONSTANT_VALUE_PARAMETER_NAMES = [
            "intValue",
            "doubleValue",
            "floatValue",
            "longValue",
            "shortValue",
            "byteValue",
            "charValue",
            "booleanValue",
            "stringValue"
    ]

    static Object getConstantFromAnnotation(AnnotationNode annotationNode) {
        def mapOfConstantValueCandidates = new HashMap<String, Object>()
        for (def constantParameterName : CONSTANT_VALUE_PARAMETER_NAMES) {
            def constantParameterValue = annotationNode.findValueByName(constantParameterName)
            if (constantParameterValue != null) {
                mapOfConstantValueCandidates[constantParameterName] = constantParameterValue
            }
        }

        if (mapOfConstantValueCandidates.isEmpty()) {
            throw new InvalidTransmutationDescriptionException("There is no constant value for annotation '${annotationNode.desc}'! Please, provide one of these parameters: ${CONSTANT_VALUE_PARAMETER_NAMES}!")
        }

        if (mapOfConstantValueCandidates.size() > 1) {
            throw new InvalidTransmutationDescriptionException("There is more than one values provided for constant annotation ${annotationNode.desc}: ${mapOfConstantValueCandidates}. The constant may has only one value! Please, choose it and remove others!")
        }

        return mapOfConstantValueCandidates.values().first()
    }

}
