package io.tswf.groovy.bettergroovy.asm.transmutation.util

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.MethodDescription
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidMethodDescriptionException
import org.objectweb.asm.Type
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
class MethodDescriptionUtil {

    static Method findMethodFromAnnotation(AnnotationNode annotationNode) {
        def name = annotationNode.getValueByName("name", String)
        def desc = annotationNode.findValueByName("descriptor", String)
        def returnTypeAsReference = annotationNode.findValueByName("returnType", AnnotationNode)
        def returnTypeAsClass = annotationNode.findValueByName("returnTypeClass", Type)
        def parametersAsReferences = annotationNode.findValueByName("parameterTypes", AnnotationNode[])
        def parametersAsClasses = annotationNode.findValueByName("parameterTypeClasses", Type[])

        // Проверяем, что пользователь не задублировал данные

        if (desc != null && (((returnTypeAsClass != null || returnTypeAsReference != null) || (parametersAsClasses != null || parametersAsReferences != null)))) {
            throw new InvalidMethodDescriptionException("Descriptor can not be used with parameter types or return type manual seclection in @${MethodDescription.class.name}")
        }

        if (returnTypeAsClass != null && returnTypeAsReference != null) {
            throw new InvalidMethodDescriptionException("Return type classes and references can not be used together in @${MethodDescription.class.name}. You must choose a single way to provide return type!")
        }

        if (parametersAsClasses != null && parametersAsReferences != null) {
            throw new InvalidMethodDescriptionException("Parameter type classes and references can not be used together in @${MethodDescription.class.name}. You must choose a single way to provide parameters!")
        }

        // Начинаем читать значения параметров аннотации
        // Если есть дескриптор в явном виде, то пофигу на все
        if (desc.isNotNullOrBlank()) {
            return new Method(name, desc)
        }

        // Читаем возвращаемый тип сначала из референсов, а потом из классов, если там нет
        def resultReturnType = (Type) null

        if (returnTypeAsReference != null) {
            resultReturnType = ClassReferenceUtil.getTypeFromClassRefAnnotation(returnTypeAsReference)
        }

        if (returnTypeAsClass != null) {
            resultReturnType = returnTypeAsClass
        }

        if (resultReturnType == null) {
            throw new InvalidMethodDescriptionException("There is no method return type description on @${MethodDescription.class.name} annotation!")
        }

        // Читаем типы параметров сначала из референсов, а потом из классов, если там нет
        def resultParameterTypes = (Type[]) null

        if (parametersAsReferences != null) {
            resultParameterTypes = parametersAsReferences.stream()
                .map { ClassReferenceUtil.getTypeFromClassRefAnnotation(it) }
                .toArray { new Type[it] }
        }

        if (parametersAsClasses != null) {
            resultParameterTypes = parametersAsClasses
        }

        if (resultParameterTypes == null) {
            throw new InvalidMethodDescriptionException("There is no method parameters description on @${MethodDescription.class.name} annotation!")
        }

        return new Method(name, resultReturnType, resultParameterTypes)
    }
}
